/*
 * $Id: results.c,v 1.1 1998/06/15 04:16:37 elkner Exp $
 *
 * Author:  Jens Elkner            elkner@ivs.cs.uni-magdeburg.de
 * Project: Jesalfa     http://ivs.cs.uni-magdeburg.de/~elkner/webtools/jesalfa/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html or ./gpl.html
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "stdlib.h"
#include "stdio.h"
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "main.h"
#include "version.h"
#include "debug.h"
#include "tree.h"
#include "ip_list.h"
#include "utils.h"
#include "fparse.h"
#include "register.h"
#include "results.h"

static void AllocTLDstack(t_tld_stack_entry ***, long);
static void AllocSLDstack(t_sld ***, long );
static int InitTLDSLDstack(t_tld *);
static int InitSLDstack(tree **);
static int InitSLDstackentry(t_sld *);
static void CalcOverallTLD(t_tld_stack_entry **, long);
static void CalcOverallSLD(t_sld **, long);
static void SortTLD(t_tld_stack_entry ***, t_tld_stack_entry ***,
		    t_tld_stack_entry **, long);
static void SortSLD(t_sld ***, t_sld ***, t_sld**, long);
static int CompareTLDByHits(t_tld_stack_entry **, t_tld_stack_entry **);
static int CompareTLDByBytes(t_tld_stack_entry **, t_tld_stack_entry **);
static int CompareSLDByHits(t_sld **, t_sld **);
static int CompareSLDByBytes(t_sld **, t_sld **);
static void PrintDomainProlog(int, FILE *, char *, int, stat_type);
static void PrintTLD(int, int, int, FILE *, stat_type, ip_acl *);
static void PrintClasses(int, FILE *);
static int PrintLogtype(t_logtype *);
static void PrintTblHead(int, int, int, FILE *,char *, char *, t_hitsize *);
static void PrintTLDTblBody(int,int,int, int, char *, char *, FILE *, stat_type);
static void PrintFooter(int, FILE *, ip_acl *);
static void PrintSLD(int, int, int, FILE *, stat_type, ip_acl *);
static void PrintSLDTblBody(int,int,int, int, char *, char *, FILE *, stat_type);
static void DeallocTLDstack(t_tld_stack_entry ***, t_tld_stack_entry ***,
			    t_tld_stack_entry ***, long);
static void DeallocSLDstack(t_sld ***, t_sld ***, t_sld ***, long);
     
static t_tld_stack_entry **tldstack, **tldbyhits, **tldbybytes;
static t_sld **sldstack, **sldbyhits, **sldbybytes;

static t_tld_stack_entry **ctldstack, **ctldbyhits, **ctldbybytes;
static t_sld **csldstack, **csldbyhits, **csldbybytes;

static t_sld **xsldstack, **xsldbyhits, **xsldbybytes;

static t_sld **psldstack, **psldbyhits, **psldbybytes;

static FILE *gfd;
static int tldidx;
static int sldidx;
static int count;
static stat_type stats;

int
PrintResults(ip_acl *ip_list)
{
    t_wordlist *entry, *next;
    char *t;
    char buffer[1024];
    int maxcount, txt, numeric;
    FILE *fd;

    if (do_domain) {
	stats = IP;
	AllocTLDstack(&tldstack,maxtld+1);   /* +1 = overall entry */
	AllocSLDstack(&sldstack,maxsld);
	tldidx = 0; sldidx = 0;			/* VERY IMPORTANT */
	tree_trav(&tld_root,InitTLDSLDstack);
	CalcOverallTLD(tldstack, maxtld);
	SortTLD(&tldbyhits,&tldbybytes,tldstack,maxtld+1);
	SortSLD(&sldbyhits,&sldbybytes,sldstack,maxsld);
    }
    if (do_content) {
	stats = CONTENT;
	AllocTLDstack(&ctldstack,maxctld+1); /* +1 = overall entry */
	AllocSLDstack(&csldstack,maxcsld);
	tldidx = 0; sldidx = 0;			/* VERY IMPORTANT */
	tree_trav(&ctld_root,InitTLDSLDstack);
	CalcOverallTLD(ctldstack,maxctld);
	SortTLD(&ctldbyhits,&ctldbybytes,ctldstack,maxctld+1);
	SortSLD(&csldbyhits,&csldbybytes,csldstack,maxcsld);
    }
    if (do_extension) {
	stats = EXT;
	AllocSLDstack(&xsldstack,maxxsld+1);
	sldidx = 0; 				/* VERY IMPORTANT */
	InitSLDstack(&xsld_root);
	xsldstack[maxxsld] = NewSLD("all", EXT);
	CalcOverallSLD(xsldstack,maxxsld);
	SortSLD(&xsldbyhits,&xsldbybytes,xsldstack,maxxsld);
    }
    if (do_proto) {
	stats = PROTO;
	AllocSLDstack(&psldstack,maxpsld+1);
	sldidx = 0; 				/* VERY IMPORTANT */
	InitSLDstack(&psld_root);
	psldstack[maxpsld] = NewSLD("all", PROTO);
	CalcOverallSLD(psldstack,maxpsld);
	SortSLD(&psldbyhits,&psldbybytes,psldstack,maxpsld);
    }
    entry = outfilelist;
    for (; entry; entry = next) {
	/* we already validated the entry in fparse.c, so no further
	   checks */
	t = strchr(entry->key,'=');
	*t = '\0';
	if ( outpath ) {
	    strcpy(buffer,outpath);
	    strcat(buffer,"/");
	    strcat(buffer,entry->key);
	}
	else
	    strcpy(buffer,entry->key);
	printf("%s\n",buffer);
	fd = fopen(buffer, "w");
	*t = '=';
	txt = ( *(t+1) == 't' ) ? 1 : 0;
	numeric = ( *(t+5) == 'n') ? 1 : 0;
	maxcount = atoi(t+7);
	switch ( *(t+3) ) {
	    case 't' :
		PrintTLD(maxcount, txt, numeric, fd, IP, ip_list);
		break;
	    case 's' :
		PrintSLD(maxcount, txt, numeric, fd, IP, ip_list);
		break;
	    case 'C' :
		PrintTLD(maxcount, txt, numeric, fd, CONTENT, ip_list);
		break;
	    case 'S' :
		PrintSLD(maxcount, txt, numeric, fd, CONTENT, ip_list);
		break;
	    case 'e' :
		PrintSLD(maxcount, txt, numeric, fd, EXT, ip_list);
		break;
	    case 'p' :
		PrintSLD(maxcount, txt, numeric, fd, PROTO, ip_list);
		break;
	    default :
		;
	}
	next = entry->next;
    }
    if (do_domain) {
	DeallocSLDstack(&sldstack,&sldbyhits,&sldbybytes,maxsld);
	DeallocTLDstack(&tldstack,&tldbyhits,&tldbybytes,maxtld+1);
    }
    if (do_content) {
	DeallocSLDstack(&csldstack,&csldbyhits,&csldbybytes,maxcsld);
	DeallocTLDstack(&ctldstack,&ctldbyhits,&ctldbybytes,maxctld+1);
    }
    if (do_extension) {
	DeleteSLD(xsldstack[maxxsld]);
	DeallocSLDstack(&xsldstack,&xsldbyhits,&xsldbybytes,maxxsld);
	memdel(sizeof(t_sld *), "Last xSLD stack entry pointer");
    }
    if (do_proto) {
	DeleteSLD(psldstack[maxpsld]);
	DeallocSLDstack(&psldstack,&psldbyhits,&psldbybytes,maxpsld);
	memdel(sizeof(t_sld *), "Last pSLD stack entry pointer");
    }
    return 0;
}

static void
SortTLD(t_tld_stack_entry ***thehits, t_tld_stack_entry ***thebytes,
	t_tld_stack_entry **stack, long entries )
{
    int i;
    t_tld_stack_entry **hits, **bytes;
    
    hits = (t_tld_stack_entry **) calloc(entries,
					 sizeof(t_tld_stack_entry *));
    memcpy(hits, stack, entries * sizeof(t_tld_stack_entry *));
    memadd(entries * sizeof(t_tld_stack_entry *),"TLD stack by Hits");
    
    bytes = (t_tld_stack_entry **)
        calloc(entries,sizeof(t_tld_stack_entry *));
    memcpy(bytes, stack, entries * sizeof(t_tld_stack_entry *));
    memadd(entries * sizeof(t_tld_stack_entry *),"TLD stack by Bytes");
#if USE_GNU
    gnuqsort(hits, entries -1, sizeof(t_tld_stack_entry *),
	     (QS) CompareTLDByHits);
    gnuqsort(bytes, entries -1, sizeof(t_tld_stack_entry *),
	     (QS) CompareTLDByBytes);
#else
    qsort(hits, entries -1, sizeof(t_tld_stack_entry *),
	  (QS) CompareTLDByHits);
    qsort(bytes, entries -1, sizeof(t_tld_stack_entry *),
	  (QS) CompareTLDByBytes);
#endif
    *thehits = hits;
    *thebytes = bytes;
}

static void
SortSLD(t_sld ***hits, t_sld ***bytes, t_sld **stack, long entries)
{
    int i;
    *hits = (t_sld **) calloc(entries,sizeof(t_sld *));
    memcpy(*hits, stack, entries * sizeof(t_sld *));
    memadd(entries * sizeof(t_sld *),"SLD stack by Hits");
    *bytes= (t_sld **) calloc(entries,sizeof(t_sld *));
    memcpy(*bytes, stack, entries * sizeof(t_sld *));
    memadd(entries * sizeof(t_sld *),"SLD stack by Bytes");
#if USE_GNU
    gnuqsort(*hits, entries, sizeof(t_sld *), (QS) CompareSLDByHits);
    gnuqsort(*bytes, entries, sizeof(t_sld *), (QS) CompareSLDByBytes);
#else
    qsort(*hits, entries, sizeof(t_sld *), (QS) CompareSLDByHits);
    qsort(*bytes, entries, sizeof(t_sld *), (QS) CompareSLDByBytes);
#endif
}

static int
CompareTLDByHits(t_tld_stack_entry **tld1, t_tld_stack_entry **tld2)
{
    t_hitsize *h1, *h2;

    if (!tld1 || !tld2) {
	fprintf(errfd,"Error in qsorting TLD stack area 1\n");
	abort();
    }
    h1 = (*tld1)->hitsize;
    h2 = (*tld2)->hitsize;
    if ( h1[maxlogtypegrpid].no > h2[maxlogtypegrpid].no )
	return (-1);
    if ( h1[maxlogtypegrpid].no < h2[maxlogtypegrpid].no )
	return (1);
    return (strcasecmp((*tld1)->name,(*tld2)->name));
}

static int
CompareSLDByHits(t_sld ** sld1, t_sld ** sld2)
{
    t_hitsize *h1, *h2;

    if (!sld1 || !sld2) {
	fprintf(errfd,"Error in qsorting SLD stack area 1\n");
	abort();
    }
    h1 = (*sld1)->hitsize;
    h2 = (*sld2)->hitsize;
    if ( h1[maxlogtypegrpid].no > h2[maxlogtypegrpid].no )
	return (-1);
    if ( h1[maxlogtypegrpid].no < h2[maxlogtypegrpid].no )
	return (1);
    return (strcasecmp((*sld1)->name,(*sld2)->name));
}

static int
CompareTLDByBytes(t_tld_stack_entry ** tld1, t_tld_stack_entry ** tld2)
{
    t_hitsize *h1, *h2;

    if (!tld1 || !tld2) {
	fprintf(errfd,"Error in qsorting TLD stack area 2\n");
	abort();
    }
    h1 = (*tld1)->hitsize;
    h2 = (*tld2)->hitsize;
    if ( h1[maxlogtypegrpid].bytes > h2[maxlogtypegrpid].bytes )
	return (-1);
    if ( h1[maxlogtypegrpid].bytes < h2[maxlogtypegrpid].bytes )
	return (1);
    return (strcasecmp((*tld1)->name,(*tld2)->name));
}

static int
CompareSLDByBytes(t_sld ** sld1, t_sld ** sld2)
{
    t_hitsize *h1, *h2;

    if (!sld1 || !sld2) {
	fprintf(errfd,"Error in qsorting SLD stack area 2\n");
	abort();
    }
    h1 = (*sld1)->hitsize;
    h2 = (*sld2)->hitsize;
    if ( h1[maxlogtypegrpid].bytes > h2[maxlogtypegrpid].bytes )
	return (-1);
    if ( h1[maxlogtypegrpid].bytes < h2[maxlogtypegrpid].bytes )
	return (1);
    return (strcasecmp((*sld1)->name,(*sld2)->name));
}

static void
CalcOverallSLD(t_sld **stack, long entries) {
    int i, k;
    t_hitsize *overall, *hit;
    
    overall = stack[entries]->hitsize;
    for(i=0; i < entries; i++) {
	hit = stack[i]->hitsize;
	for(k=0; k < maxlogtypegrpid; k++) {
	    overall[k].no += hit[k].no;
	    overall[k].bytes += hit[k].bytes;
	}
    }
    for(k=0; k < maxlogtypegrpid; k++) {
	overall[maxlogtypegrpid].no += overall[k].no;
	overall[maxlogtypegrpid].bytes += overall[k].bytes;
    }	
}

static void
CalcOverallTLD(t_tld_stack_entry **stack, long entries) {
    int i, k;
    t_hitsize *hit, *overall;
    
    overall = stack[entries]->hitsize;
    for(i=0; i < entries; i++) {
	hit = stack[i]->hitsize;
	for(k=0; k < maxlogtypegrpid; k++) {
	    overall[k].no += hit[k].no;
	    hit[maxlogtypegrpid].no += hit[k].no;
	    overall[k].bytes += hit[k].bytes;
	    hit[maxlogtypegrpid].bytes += hit[k].bytes;
	}
    }
    for(k=0; k < maxlogtypegrpid; k++) {
	overall[maxlogtypegrpid].no += overall[k].no;
	overall[maxlogtypegrpid].bytes += overall[k].bytes;
    }
}

static void
PrintTLD(int maxcount, int txt, int numeric, FILE *fd, stat_type s,
	 ip_acl *ip_list)
{
    char fmthitsize[64];
    char fmtdomain[64];
    char title[64];
    char what[64];
    char abbrev[12];
    long entries;
    t_hitsize *overall;

    if (s == IP) {
	overall = tldstack[maxtld]->hitsize;
	entries = maxtld;
	strcpy(abbrev,"TLD");
	strcpy(what,"TopLevelDomains");
    }
    else {
	overall = ctldstack[maxctld]->hitsize;
	entries = maxctld;
	strcpy(abbrev,"CT");
	strcpy(what,"ContentTypes");
    }
    sprintf(title,"Access Log Statistics (%s)",abbrev);
    if ( (maxcount < 1) || (maxcount > entries) )
	maxcount = entries;
    if (txt) {
	strcpy(fmtdomain,"%s:");
	if (numeric)
	    strcpy(fmthitsize, " %d/%lld");
	else
	    strcpy(fmthitsize, " %1.1f/%1.1f");
    }
    else {
	strcpy(fmtdomain,"<TR>\n<TH>%s ");
	if (numeric)
	    strcpy(fmthitsize,"<TD ALIGN=RIGHT>%s <TD ALIGN=RIGHT>%s ");
	else
	    strcpy(fmthitsize,"<TD ALIGN=RIGHT>%1.1f <TD ALIGN=RIGHT>%1.1f ");
    }
    PrintDomainProlog(txt, fd, title, 0, s);
    PrintClasses(txt, fd);
    PrintTblHead(txt, 0, numeric, fd, what, abbrev, overall);
    PrintTLDTblBody(txt, maxcount, numeric, 0, fmtdomain, fmthitsize, fd, s);
    if (txt)
	fprintf(fd,"\n");
    else
	fprintf(fd,"</TABLE>\n<P>\n");
    PrintTblHead(txt, 1, numeric, fd, what, abbrev, overall);
    PrintTLDTblBody(txt, maxcount, numeric, 1, fmtdomain, fmthitsize, fd, s);
    if (txt)
	fprintf(fd,"\n");
    else
	fprintf(fd,"</TABLE>\n<P>\n");
    PrintFooter(txt, fd, ip_list);
}

static void
PrintSLD(int maxcount, int txt, int numeric, FILE *fd, stat_type s,
	 ip_acl *ip_list)
{
    char fmthitsize[64];
    char fmtdomain[64];
    char what[64];
    char title[64];
    char abbrev[12];
    long entries;
    char delimiter[2];
    t_hitsize *overall;

    if ( s == IP ) {
	overall = tldstack[maxtld]->hitsize;
	entries = maxsld;
	strcpy(abbrev,"SLD");
	strcpy(what,"SecondLevelDomains");
	strcpy(delimiter,".");
    }
    else if ( s == CONTENT ) {
	overall = ctldstack[maxctld]->hitsize;
	entries = maxcsld;
	strcpy(abbrev,"CST/CT");
	strcpy(what,"ContentSubTypes");
	strcpy(delimiter,"/");
    }
    else if ( s == EXT ) {
	overall = xsldstack[maxxsld]->hitsize;
	entries = maxxsld;
	strcpy(abbrev,"EXT");
	strcpy(what,"Extensions (case-insensitive)");
	strcpy(delimiter,"");
    }
    else { /* s == PROTO */
	overall = psldstack[maxpsld]->hitsize;
	entries = maxpsld;
	strcpy(abbrev,"PROTO");
	strcpy(what,"Protocols");
	strcpy(delimiter,"");
    }
    sprintf(title,"Access Log Statistics (%s)",abbrev);
    if ( (maxcount < 1) || (maxcount > entries) )
	maxcount = entries;
    if (txt) {
	sprintf(fmtdomain,"%s%s%s","%s",delimiter,"%s:");
	if (numeric)
	    strcpy(fmthitsize, " %d/%lld");
	else
	    strcpy(fmthitsize, " %1.1f/%1.1f");
    }
    else {
	sprintf(fmtdomain,"<TR>\n<TH>%s%s%s ","%s",delimiter,"%s");
	if (numeric)
	    strcpy(fmthitsize,"<TD ALIGN=RIGHT>%s <TD ALIGN=RIGHT>%s ");
	else
	    strcpy(fmthitsize,"<TD ALIGN=RIGHT>%1.1f <TD ALIGN=RIGHT>%1.1f ");
    }
    PrintDomainProlog(txt, fd, title, 0, s);
    PrintClasses(txt, fd);
    /* by hits */
    PrintTblHead(txt, 0, numeric, fd, what, abbrev, overall);
    PrintSLDTblBody(txt, maxcount, numeric, 0, fmtdomain, fmthitsize, fd, s);
    if (txt)
	fprintf(fd,"\n");
    else
	fprintf(fd,"</TABLE>\n<P>\n");
    /* by bytes */
    PrintTblHead(txt, 1, numeric, fd, what, abbrev, overall);
    PrintSLDTblBody(txt, maxcount, numeric, 1, fmtdomain, fmthitsize, fd, s);
    if (txt)
	fprintf(fd,"\n");
    else
	fprintf(fd,"</TABLE>\n<P>\n");
    PrintFooter(txt, fd, ip_list);
}

static void
PrintClasses(int txt, FILE *fd)
{
    if (txt)
	fprintf(fd,"%s\n=================================\n\n",
		"Logtype Classes");
    else
	fprintf(fd,
		"<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=3 WIDTH=100%%>\n"
		"<TR><TD BGCOLOR=\"#f0f0ff\" WIDTH=100%% ALIGN=CENTER>\n"
		"<FONT COLOR=black SIZE=+1><STRONG>\n"
		"Logtype Classes\n"
		"</FONT></TABLE>\n<P>\n"
		"<TABLE BORDER=0>\n<TR>\n");
    /* I hate global vars (count, gfd), but I don't want to make the tree
       stuff to complicated by passing unions etc. ... */
    gfd = fd;
    for(count=0; count < maxlogtypegrpid; count++) {
	if (txt)
	    fprintf(fd,"%2d ... {",count+1);
	else
	    fprintf(fd,"<TR>\n<TH VALIGN=TOP><FONT SIZE=-1>%d</FONT><TD VALIGN=TOP><FONT SIZE=-1>...</FONT><TD VALIGN=TOP><FONT SIZE=-1>",count+1);
	tree_trav(&logtype_root,PrintLogtype);
	if (txt)
	    fprintf(fd," }\n");
	else
	    fprintf(fd,"</FONT>\n");
    }
    gfd = NULL;
    if (txt)
	fprintf(fd,"\n");
    else
	fprintf(fd,"</TABLE>\n<P>\n");
}

static int
PrintLogtype(t_logtype *lt)
{
    if (count == lt->id)
	fprintf(gfd," %s", lt->name);
    return TRUE;
}

static void
PrintDomainProlog(int txt, FILE *fd, char *title, int what, stat_type s)
{
    t_hitsize *overall;
    t_hitsize tcp;
    char tmp[9][64];
    long average;

    if (what == 0) {
	if ( s == IP )
	    overall = tldstack[maxtld]->hitsize;
	else if ( s == CONTENT)
	    overall = ctldstack[maxctld]->hitsize;
	else if ( s == EXT )
	    overall = xsldstack[maxxsld]->hitsize;
	else /* s == PROTO */
	    overall = psldstack[maxpsld]->hitsize;
    }
    tcp.no = overall[maxlogtypegrpid].no - udp.no;
    tcp.bytes = overall[maxlogtypegrpid].bytes - udp.bytes;
    average = overall[maxlogtypegrpid].no ? overall[maxlogtypegrpid].bytes/overall[maxlogtypegrpid].no : 0 ;
    if (txt) {
	fprintf(fd,"%s\n=================================\n\n", title);
	fprintf(fd,"Counted Hits (TCP/UDP): %d (%d/%d)\n",
		overall[maxlogtypegrpid].no, tcp.no, udp.no);
	fprintf(fd,"X-fered Bytes (TCP/UDP): %lld (%lld/%lld)\n",
		overall[maxlogtypegrpid].bytes, tcp.bytes, udp.bytes);
	fprintf(fd,"Average Object Size in Bytes (TCP/UDP): %ld (%lld/%lld)\n\n",
		average,
		tcp.no ? tcp.bytes/tcp.no : 0,
		udp.no ? udp.bytes/udp.no : 0);
    }
    else {
	Num2Str(overall[maxlogtypegrpid].no,tmp[0]);
	Num2Str(tcp.no,tmp[1]);
	Num2Str(udp.no,tmp[2]);
	Num2Str(overall[maxlogtypegrpid].bytes,tmp[3]);
	Num2Str(tcp.bytes,tmp[4]);
	Num2Str(udp.bytes,tmp[5]);
	Num2Str(average,tmp[6]);
	if ( tcp.no )
	    Num2Str(tcp.bytes/tcp.no,tmp[7]);
	else
	    Num2Str(0,tmp[7]);
	if ( udp.no )
	    Num2Str(udp.bytes/udp.no,tmp[8]);
	else
	    Num2Str(0,tmp[8]);
	fprintf(fd,"<HTML>\n<HEAD>\n<TITLE>%s</TITLE>\n<HEAD>\n", title);
	fprintf(fd,"<BODY BGCOLOR=WHITE TEXT=Black>\n"
		"<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=3 WIDTH=100%%>\n"
		"<TR><TD BGCOLOR=\"#1D098E\" WIDTH=100%% ALIGN=CENTER>\n"
		"<FONT COLOR=white SIZE=+2><STRONG>\n");
	fprintf(fd,"%s\n</FONT></TABLE>\n<P>\n", title);
	fprintf(fd,"<TABLE BORDER=0>\n<TR><TH>&nbsp;<TH>overall <TH>TCP <TH>UDP\n");
	fprintf(fd,"<TR><TH>Counted Hits: <TD align=center>%s <TD align=center>%s <TD align=center>%s\n",
		tmp[0], tmp[1], tmp[2]);
	fprintf(fd,"<TR><TH>X-fered Bytes: <TD align=center>%s <TD align=center>%s <TD align=center>%s\n",
		tmp[3], tmp[4], tmp[5]);
	fprintf(fd,"<TR><TH>Average Object Size in Bytes: <TD align=center>%s <TD align=center>%s <TD align=center>%s\n</TABLE>\n<P>\n",
		tmp[6], tmp[7], tmp[8]);
    }
}

static void
PrintTblHead(int txt, int byhits, int num, FILE *fd, char *ld, char *abbrev,
	     t_hitsize *overall)
{
    int i;
    
    if (txt) {
	fprintf(fd,"%s by %s\n", ld, byhits ? "Hits" : "Bytes");
	fprintf(fd,"===========================================\n");
    }
    else {
	fprintf(fd,
		"<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=3 WIDTH=100%%>\n"
		"<TR><TD BGCOLOR=\"#f0f0ff\" WIDTH=100%% ALIGN=CENTER>\n"
		"<FONT COLOR=black SIZE=+1><STRONG>\n");
	fprintf(fd,"%s by %s\n", ld, byhits ? "Hits" : "Bytes");
	fprintf(fd,"</FONT></TABLE>\n<P>\n");
	
	fprintf(fd,"<TABLE COLS=%d BORDER=1>\n",maxlogtypegrpid*2 + 1);
	fprintf(fd, "<CAPTION ALIGN=BOTTOM> all entries in %s</CAPTION>\n"
		"<TR>\n<TH ALIGN=RIGHT>CLASS", num ? "BYTES" : "PERCENT");
	for(i=0; i < maxlogtypegrpid; i++)
	    if ( overall[i].no )
		fprintf(fd,"<TH COLSPAN=2 ALIGN=CENTER>%d ",i+1);
	fprintf(fd,"<TH COLSPAN=2 ALIGN=CENTER>overall"
		"<TH ALIGN=CENTER>av.&nbsp;Size"
		"<TR>\n<TH ALIGN=LEFT>%s", abbrev);
	for(i=0; i <= maxlogtypegrpid; i++)
	    if ( overall[i].no )
		fprintf(fd,"<TH ALIGN=CENTER>Hits<TH ALIGN=CENTER>Bytes");
	fprintf(fd,"<TH ALIGN=CENTER>Bytes");
    }
    fprintf(fd,"\n");
}

static void
PrintTLDTblBody(int txt, int max, int num, int byhits,
		char *fmtd, char *fmth, FILE *fd, stat_type s)
{
    int i, k;
    t_tld_stack_entry **x;
    t_hitsize *overall, *hit;
    int done = 0;
    int maxcount;
    long av;
    char tmp[32], tmp2[32];

    if ( s == IP ) {
	overall = tldstack[maxtld]->hitsize;
	x = byhits ? tldbyhits : tldbybytes;
	maxcount = maxtld;
    }
    else {
	overall = ctldstack[maxctld]->hitsize;
	x = byhits ? ctldbyhits : ctldbybytes;
	maxcount = maxctld;
    }
    if ( overall[maxlogtypegrpid].no == 0 ) {
	fprintf(errfd,"No hits at all - nothing to print\n");
	return;
    }
    if ( overall[maxlogtypegrpid].bytes == 0) {
	fprintf(errfd,"Nothing transfered at all - nothing to print\n");
	return;
    }	
    for(i=0; i < max; i++) {
last:	hit = x[i]->hitsize;
	fprintf(fd,fmtd,x[i]->name);
	if (num) {
	    for(k=0; k <= maxlogtypegrpid; k++)
		if (txt)
		    fprintf(fd, fmth, hit[k].no, hit[k].bytes);
		else if (overall[k].no) {
		    Num2Str(hit[k].no,tmp);
		    Num2Str(hit[k].bytes,tmp2);
		    fprintf(fd, fmth, tmp, tmp2);
		}
	}
	else {
	    for(k=0; k <= maxlogtypegrpid; k++) {
		if ( txt || overall[k].no )
		    fprintf(fd, fmth,
			    100.0 * hit[k].no/overall[maxlogtypegrpid].no,
			    100.0 * hit[k].bytes/overall[maxlogtypegrpid].bytes);
	    }
	}
	av = hit[maxlogtypegrpid].bytes/hit[maxlogtypegrpid].no;
	if (txt)
	    fprintf(fd," %ld",av);
	else {
	    Num2Str(av,tmp);
	    fprintf(fd,"<TD ALIGN=RIGHT> %s ",tmp);
	}
	fprintf(fd, "\n");
    }
    if ( ! done ) {
	done++;
	i = maxcount;
	max = maxcount;
	goto last;
    }
}

static void
PrintSLDTblBody(int txt, int max, int num, int byhits, char *fmtd, char *fmth,
		FILE *fd, stat_type s)
{
    int i, k;
    t_sld **x;
    t_hitsize *overall, *hit;
    char *name;
    int maxcount;
    long av;
    char tmp[32], tmp2[32];
    
    if ( s == IP ) {
	overall = tldstack[maxtld]->hitsize;
	x = byhits ? sldbyhits : sldbybytes;
	name = tldstack[maxtld]->name;
    }
    else if ( s == CONTENT ) {
	overall = ctldstack[maxctld]->hitsize;
	x = byhits ? csldbyhits : csldbybytes;
	name = ctldstack[maxctld]->name;
    }
    else if ( s == EXT ) {
	overall = xsldstack[maxxsld]->hitsize;
	x = byhits ? xsldbyhits : xsldbybytes;
	name = xsldstack[maxxsld]->name;
    }
    else { /* s == PROTO */
	overall = psldstack[maxpsld]->hitsize;
	x = byhits ? psldbyhits : psldbybytes;
	name = psldstack[maxpsld]->name;
    }
    if ( overall[maxlogtypegrpid].no == 0 ) {
	fprintf(errfd,"No hits at all - nothing to print\n");
	return;
    }
    if ( overall[maxlogtypegrpid].bytes == 0) {
	fprintf(errfd,"Nothing transfered at all - nothing to print\n");
	return;
    }
    for(i=0; i < max; i++) {
	hit = x[i]->hitsize;
	fprintf(fd,fmtd,x[i]->name, x[i]->tld_name ? x[i]->tld_name : "");
	if (num) {
	    for(k=0; k <= maxlogtypegrpid; k++)
		if ( txt)
		    fprintf(fd, fmth, hit[k].no, hit[k].bytes);
		else if (overall[k].no ) {
		    Num2Str(hit[k].no, tmp);
		    Num2Str(hit[k].bytes, tmp2);
		    fprintf(fd, fmth, tmp, tmp2);
		}
	}
	else {
	    for(k=0; k <= maxlogtypegrpid; k++) {
		if ( txt || overall[k].no )
		    fprintf(fd, fmth,
			    100.0 * hit[k].no/overall[maxlogtypegrpid].no,
			    100.0 * hit[k].bytes/overall[maxlogtypegrpid].bytes);
	    }
	}
	av = hit[maxlogtypegrpid].bytes/hit[maxlogtypegrpid].no;
	if (txt)
	    fprintf(fd," %ld",av);
	else {
	    Num2Str(av,tmp);
	    fprintf(fd,"<TD ALIGN=RIGHT> %s ",tmp);
	}
	fprintf(fd, "\n");
    }
    /* overall */
    fprintf(fd,fmtd,name, "");
    if (num) {
	for(k=0; k <= maxlogtypegrpid; k++) {
	    if (txt)
		fprintf(fd, fmth, overall[k].no, overall[k].bytes);
	    else if (overall[k].no ) {
		Num2Str(overall[k].no, tmp);
		Num2Str(overall[k].bytes, tmp2);
		fprintf(fd, fmth, tmp, tmp2);
	    }
	}
    }
    else {
	for(k=0; k <= maxlogtypegrpid; k++) {
	    if ( txt || overall[k].no )
		fprintf(fd, fmth,
			100.0 * overall[k].no/overall[maxlogtypegrpid].no,
			100.0 * overall[k].bytes/overall[maxlogtypegrpid].bytes);
	}
    }
    av = overall[maxlogtypegrpid].bytes/overall[maxlogtypegrpid].no;
    if (txt)
	fprintf(fd," %ld",av);
    else {
	Num2Str(av,tmp);
	fprintf(fd,"<TD ALIGN=RIGHT> %s ",tmp);
    }
    fprintf(fd, "\n");
}

static void
PrintFooter(int txt, FILE *fd, ip_acl *ip_list)
{
    char date[64];
    ip_acl *ip;
    unsigned long tmp;
    
    GetLocalAsciiTime(date);
    if ( ip_list ) {
	if (! txt)
	    fprintf(fd,"<!--\n");
	fprintf(fd, "IP exclude patterns: ");
	for(ip=ip_list; ip; ip=ip->next )
	    fprintf(fd, "%s%s/%d ", ip->access == IP_ALLOW ? "!" : "",
		    inet_ntoa(ip->addr), cidr_ntod(ip->mask));
	fprintf(fd, "\n");
	if (txt)
	    fprintf(fd,"\n");
	else
	    fprintf(fd,"-->\n\n");
    }
    if (protolist) {
	    t_wordlist *p;
	    if (! txt)
		fprintf(fd,"<!--\n");
	    fprintf(fd, "Protocols excluded: ");
	    for(p=protolist; p; p=p->next)
		fprintf(fd,"%s ", p->key ? p->key : "");
	    fprintf(fd,"\n");
	    if (txt)
		fprintf(fd,"\n");
	    else
		fprintf(fd, "-->\n\n");
    }
    if (txt)
	fprintf(fd,"Generated by %s %s    (C) by %s (%s)\n\n"
		"Last modified: %s\n\n%s\n",
		APPNAME, VERSION, AUTHOR, AUTHOR_EMAIL, date, APPURL);
    else
	fprintf(fd,
		"<P>"
		"<TABLE Width=100%%>\n"
		"<TR><TD BGCOLOR=\"#F0F0F0\"><FONT SIZE=-1>\n"
		"Generated by <A HREF=\"%s\">%s %s</A> &nbsp;&nbsp\n"
		"(C) by <A HREF=\"%s\">%s</A>&nbsp;&nbsp;"
		"<A HREF=\"mailto:%s\">&lt;%s&gt;</A>\n"
		"<TR><TD ALIGN=RIGHT><FONT SIZE=-2>"
		"Last modified: %s\n</FONT></TABLE>\n</BODY></HTML>\n",
		APPURL,APPNAME, VERSION,
		AUTHOR_URL, AUTHOR, AUTHOR_EMAIL, AUTHOR_EMAIL, date);
}

static int
InitTLDSLDstack(t_tld *tld)
{
    if ( stats == IP )
	tldstack[tldidx]->name = tld->name;
    else
	ctldstack[tldidx]->name = tld->name;
    InitSLDstack(&(tld->sld_root));
    tldidx++;
    return TRUE;
}

static int
InitSLDstack(tree **t) 
{
    if (t)
	tree_trav(t, InitSLDstackentry);
    return TRUE;
}

static int
InitSLDstackentry(t_sld *sld)
{
    /* here we need to investigate a little bit further :( */
    /* probably sld[->name] = NUL */
    t_hitsize *sldhit, *tldhit;
    int i, do_tld;

    do_tld = 1;
    if ( stats == IP ) {
	sldstack[sldidx] = sld;
	tldhit = tldstack[tldidx]->hitsize;
    }
    else if (stats == CONTENT) {
	csldstack[sldidx] = sld;
	tldhit = ctldstack[tldidx]->hitsize;
    }
    else if (stats == EXT) {
	xsldstack[sldidx] = sld;
	do_tld = 0;
    }
    else { /* stats == PROTO */
	psldstack[sldidx] = sld;
	do_tld = 0;
    }
    sldhit = sld->hitsize;
   
    for(i=0; i < maxlogtypegrpid; i++) {	
	sldhit[maxlogtypegrpid].no += sldhit[i].no;
	sldhit[maxlogtypegrpid].bytes += sldhit[i].bytes;
	if (do_tld) {
	    tldhit[i].no += sldhit[i].no;
	    tldhit[i].bytes += sldhit[i].bytes;
	}
    }
    /* if ( sld ) {
         if (sld->name)
           fprintf(errfd,"%6d %15d \"%s.%s\"\n",sldidx,sld->lineno,sld->name,sld->tld_name);
       }
    */
    sldidx++;
    return TRUE;
}

static void
AllocTLDstack(t_tld_stack_entry ***thestack, long entries)
{
    int i;
    t_tld_stack_entry **stack;
    
    stack = (t_tld_stack_entry **)
	calloc(entries, sizeof(t_tld_stack_entry *));
    memadd(entries * sizeof(t_tld_stack_entry *),"TLD stack");
    for(i=0; i < entries; i++) {
	stack[i] = (t_tld_stack_entry *) malloc(sizeof(t_tld_stack_entry));
	memadd(sizeof(t_tld_stack_entry),"  Stack entry");
	stack[i]->name = NULL;
	stack[i]->hitsize = (t_hitsize *) calloc(maxlogtypegrpid+1,
						 sizeof(t_hitsize));
	memadd((maxlogtypegrpid+1) *sizeof(t_hitsize),"    n * t_hitsize");
    }
    stack[entries-1]->name = savestr("all");
    debug(4,"TLD stack with %d entries allocated",entries);
    *thestack = stack;
}

/* order doesn't matter */
static void
DeallocTLDstack(t_tld_stack_entry ***stack1, t_tld_stack_entry ***stack2,
		t_tld_stack_entry ***stack3,long entries)
{
    int i;
    t_tld_stack_entry **stack;

    stack = *stack1;
    free(stack);
    memdel(entries * sizeof(t_tld_stack_entry *),"TLD stack 1");
    stack = *stack2;
    free(stack);
    stack = *stack3;
    memdel(entries * sizeof(t_tld_stack_entry *),"TLD stack 2");
    delstr(stack[entries -1]->name); /* delete overall tld name only -
					   all others are still part of the
					   tld tree !!! */
    for(i=0; i < entries; i++) {
	free(stack[i]->hitsize);
	memdel((maxlogtypegrpid+1) *sizeof(t_hitsize),"    n * t_hitsize");
	free(stack[i]);
	memdel(sizeof(t_tld_stack_entry),"  Stack entry");
    }
    free(stack);
    memdel(entries * sizeof(t_tld_stack_entry *),"TLD stack 3");
    debug(4,"TLD stacks with %d entries deallocated", entries);
}

static void
AllocSLDstack(t_sld ***stack, long entries)
{
    *stack = (t_sld **) calloc(entries,sizeof(t_sld *));
    memadd(entries *sizeof(t_sld *),"SLD stack");
    debug(4,"SLD stack with %d entries allocated",entries);
}

/* oder does not matter */
static void
DeallocSLDstack(t_sld ***stack1, t_sld ***stack2, t_sld ***stack3, long entries)
{
    if (*stack1) {
	free(*stack1);
	*stack1 = NULL;
	memdel(entries * sizeof(t_sld *),"SLD stack 1");
    }
    if (*stack2) {
	free(*stack2);
	*stack2 = NULL;
	memdel(entries * sizeof(t_sld *),"SLD stack 2");
    }
    if (*stack3) {
	free(*stack3);
	*stack3 = NULL;
	memdel(entries * sizeof(t_sld *),"SLD stack 3");
    }
    debug(4,"SLD stacks with %d entries deallocated",entries);
}
