/*
 * $Id: utils.c,v 1.1 1998/06/15 04:15:46 elkner Exp $
 *
 * Author:  Jens Elkner            elkner@ivs.cs.uni-magdeburg.de
 * Project: Jesalfa     http://ivs.cs.uni-magdeburg.de/~elkner/webtools/jesalfa/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html or ./gpl.html
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <errno.h>

#include "debug.h"
#include "utils.h"

char *
savestr(const char *str)
{
        char    *save;

        save = (char *)malloc(strlen(str) + 1);
	if ( save == NULL ) {
	    perror("Problems allocating memory for a string:");
	    debug(3,"Problems allocating memory for \"%s\"",str);
	}
	else {
	    strcpy(save, str);
	    memadd(strlen(save) + 1, str);
	}
        return save;
}

void
delstr(const char *str)
{
    if (!str)
	return;
    memdel(strlen(str) + 1, str);
    free(str);
}

void
memadd(int size, const char *t)
{
    memusage += size;
    if (memusage > maxmem)
	maxmem = memusage;
    debug(999,"mem alloc: %8d Bytes (+%d %s)",memusage,size,t);
}

void
memdel(int size, const char *t)
{
    memusage -= size;
    debug(999,"mem  free: %8d Bytes (-%d %s)",memusage,size,t);
}

void
GetTimeDiff(struct timeval *start,struct timeval *stop, char *buf)
{
    long sdiff, udiff;
    int hour,min;
    
    sdiff = stop->tv_sec - start->tv_sec;
    udiff = stop->tv_usec - start->tv_usec;
    if ( udiff < 0 ) {
	udiff *= -1;
	sdiff--;
    }
    hour = sdiff / (60 * 60);
    sdiff -= hour * 60 * 60;
    min = sdiff / 60;
    sdiff -= min * 60;
    udiff /= 1000;
    sprintf(buf,"%02d:%02d:%02d.%03d",hour,min,sdiff,udiff);
}

void
SetStartTime(void)
{
   gettimeofday(&starttime, NULL); 
}

void
SetStopTime(void)
{
   gettimeofday(&stoptime, NULL);
}

int
GetLocalAsciiTime(char *timebuf)
{
    time_t t;
    struct tm *tms;

    t=time(NULL);
    tms = localtime(&t);
    strftime(timebuf,64,"%A, %d-%b-%y %T %Z",tms);
    return 0;
}

void
Num2Str(unsigned long long num,char *buf2)
{
    char buf1[64];
    int i, len, remain, idx1, idx2, count;
    
    sprintf(buf1,"%lld",num);
    len = strlen(buf1);
    remain = len % 3;
    idx1 = 0; idx2 = 0;
    if (remain) {
	for(i=0; i < remain; i++)
	    buf2[idx2++] = buf1[idx1++];
	buf2[idx2++] = '.';
    }
    count = 0;
    for(i=remain; i < len; i ++) {
	buf2[idx2++] = buf1[idx1++];
	count++;
	if ( count == 3 ) {
	    buf2[idx2++] = '.';
	    count = 0;
	}
    }
    buf2[--idx2] = '\0';
    return;
}

void
DeleteWordList(t_wordlist *list)
{
    t_wordlist *tmp, *tmp1;

    tmp = list;
    while (tmp) {
	delstr(tmp->key);
	tmp1 = tmp->next;
	free(tmp);
	memdel(sizeof(t_wordlist),"Word list entry");
	tmp = tmp1;
    }
}
